#define _CRT_SECURE_NO_WARNINGS

#define TMP_GRAPH_IO

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tmp_graph.h"

#define GRAPH_BASE_REAL int

#define GRAPH_NAMESPACE GOMORY_HU_TEST
#include "bk.inc"
#undef GRAPH_NAMESPACE

int ComputeCut(TmpGraph<int>* input, int s, int t)
{
	int i, e, n = input->n;

	Buffer buf(1024);
	ReusableBuffer rbuf_common, rbuf[MAXFLOW_RBUF_NUM];
	typedef GOMORY_HU_TEST::Graph G;
	G g(n, input->m, &buf, &rbuf_common, rbuf);
	int* W = new int[n];
	memset(W, 0, n*sizeof(int));
	for (e=0; e<input->m; e++)
	{
		int i = input->edges[2*e], j = input->edges[2*e+1], w = input->weights[e];
		g.add_edge(i, j, w, w);
		W[i] += w;
		W[j] += w;
	}
	int W_max = 0;
	for (i=0; i<n; i++) { if (W_max < W[i]) W_max = W[i]; }
	W_max += 1;
	delete [] W;

	g.add_tweights(s, W_max, 0);
	g.add_tweights(t, 0, W_max);
	g.maxflow();

	int cost = 0;
	for (e=0; e<input->m; e++)
	{
		if (g.what_segment(input->edges[2*e]) != g.what_segment(input->edges[2*e+1])) cost += input->weights[e];
	}
	return cost;
}




void Usage(const char* s)
{
	printf("Usage: %s <input file> <solution file>\n", s);
	exit(1);
}




int main(int argc, char* argv[])
{
	if (argc != 3) Usage(argv[0]);
	TmpGraph<int>* g = ReadDIMACS(argv[1]);
	TmpGraph<int>* T = ReadDIMACS(argv[2]);
	
	int e;
	for (e=0; e<T->m; e++)
	{
		int s = T->edges[2*e], t = T->edges[2*e+1];
		int real_cost = ComputeCut(g, s, t);
		if (real_cost != T->weights[e])
		{
			printf("incorrect Gomory-Hu cut for (s,t)=(%d,%d): true value=%d, stored value=%d\n", s, t, real_cost, T->weights[e]);
			exit(1);
		}
	}
	printf("cut values for edges in the tree are correct\n");

	return 0;
}








