#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <algorithm>
#include "gomory_hu.h"

#ifdef _MSC_VER
	#define __int32_t __int32
#endif


// #define IBFS


#ifndef AL_PLUS
#ifndef AL
#ifndef FS_PLUS
#ifndef FS
	#define FS_PLUS
#endif
#endif
#endif
#endif

#ifndef BK
#ifndef IBFS
	#define BK
#endif
#endif

//////////////////////////////////////////////////////////////////
#ifdef AL_PLUS
	#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
#endif
#ifdef AL
	#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
	#define KEEP_REV_ARC_POINTER
#endif
#ifdef FS_PLUS
	#define GRAPH_BASE_FILE "graph_base_forward_star.h"
#endif
#ifdef FS
	#define GRAPH_BASE_FILE "graph_base_forward_star.h"
	#define DO_NOT_KEEP_SATURATION_FLAG
#endif
//////////////////////////////////////////////////////////////////

#define KEEP_STLISTS
#define GRAPH_BASE_REAL GH_REAL

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

int maxflow_num;
int merge_num;
double maxflow_time;
double maxflow_total_node_num;
double maxflow_total_edge_num;

struct TreeNode
{
	TreeNode*	parent;
	TreeNode*	first_child;
	TreeNode*	sibling;
	GH_REAL		cut_cost;
	union
	{
		TreeNode*	tmp;
		int			index;
	};
};

#define GRAPH_CONTRACTION
#define IS_GOOD_NODE(i) (!(i)->gh_fixed)

// will be part of GraphBase::node
#define EXTRA_NODE_FIELDS \
	unsigned int	gh_smaller_component : 1;\
	unsigned int	gh_fixed : 1;\
	unsigned int	gh_index;\
	union\
	{\
		__int32_t		gh_dist[2];\
		TreeNode*		gh_tree_node;\
	};


#ifdef BK

	#define GRAPH_NAMESPACE GOMORY_HU_BK
	#include "bk.inc"
	#undef GRAPH_NAMESPACE
	typedef GOMORY_HU_BK::Graph G;

#else

	#define GRAPH_NAMESPACE GOMORY_HU_IBFS
	#include "ibfs.inc"
	#undef GRAPH_NAMESPACE
	typedef GOMORY_HU_IBFS::IBFSGraph G;

#endif

void GomoryHuOptions::Print()
{
	int i;
	std::cout << "options: treshold=" << threshold << ", st_selection=";
	switch (st_selection)
	{
		case RANDOM_RANDOM:               std::cout << "RANDOM-RANDOM"; break;
		case RANDOM_CLOSEST:              std::cout << "RANDOM-CLOSEST"; break;
		case HEAVIEST_HEAVIEST:           std::cout << "HEAVIEST-HEAVIEST"; break;
		case HEAVIEST_HEAVIEST_EFFICIENT: std::cout << "HEAVIEST-HEAVIEST-EFFICIENT"; break;
		case HEAVIEST_CLOSEST:            std::cout << "HEAVIEST-CLOSEST"; break;
		case CLOSEST_PAIR:                std::cout << "CLOSEST-PAIR"; break;
		default: std::cout << "UNKONWN!\n"; exit(1);
	};
	
	if (st_freq > 0) std::cout << ", st_freq=" << st_freq;
#ifdef BK
	std::cout << ", BK algorithm";
#else
	std::cout << ", IBFS algorithm";
#endif
#ifdef GRAPH_BASE_FORWARD_STAR
	std::cout << " (forward star";
	#ifdef KEEP_SATURATION_FLAG
		std::cout << " +";
	#endif
#else
	std::cout << " (adjacency list";
	#ifndef KEEP_REV_ARC_POINTER
		std::cout << " +";
	#endif
#endif
	std::cout << ")";
	if (method == METHOD_T)
	{
		std::cout << ", method=T";
		for (i=0; i<stage_num; i++) std::cout << ":" << T_array[i];
	}
	if (method == METHOD_D)
	{
		std::cout << ", method=D";
		for (i=0; i<stage_num; i++) std::cout << ":(" << D_array[2*i] << "," << D_array[2*i+1] << ")";
	}
	std::cout << "\n";
}

///////////////////////////////////////////////////


struct Tree
{
	Tree(int node_num_max) : n(0), n_max(node_num_max)
	{
		nodes = new TreeNode[n_max];
		memset(nodes, 0, n_max*sizeof(TreeNode));
	}
	~Tree()
	{
		delete [] nodes;
	}
	TreeNode* NewNode()
	{
		assert(n < n_max);
		return &nodes[n ++];
	}
	void ChangeRoot(TreeNode* r)
	{
		TreeNode* i = r;
		while ( i->parent )
		{
			i->parent->tmp = i;
			i = i->parent;
		}
		while ( i != r )
		{
			assert(i->sibling == NULL);
			TreeNode* c = i->parent = i->tmp;
			i->tmp = c->parent = NULL;
			TreeNode** ptr = &i->first_child;
			while (*ptr != c) ptr = &((*ptr)->sibling);
			*ptr = c->sibling;
			c->sibling = NULL;
			i->sibling = c->first_child;
			c->first_child = i;
			i->cut_cost = c->cut_cost;
			c->cut_cost = 0;
			i = c;
		}
	}

	int n, n_max;
	TreeNode* nodes;

	GomoryHuOptions	options;

	ReusableBuffer common_rbuf;
};

const int BUFFER_LIST_BLOCK_SIZE = 256;

struct BufferList
{
	BufferList() : buf(BUFFER_LIST_BLOCK_SIZE), next(NULL) {}
	~BufferList() { if (next) delete next; }
	void* Alloc(int size) { return buf.Alloc(size); }
	BufferList* GetNext() { if (!next) next = new BufferList; return next; } 

	Buffer buf;
	ReusableBuffer rbuf[MAXFLOW_RBUF_NUM+1];
	BufferList* next;
};

class GomoryHu
{
public:
	GomoryHu(TmpGraph<GH_REAL>* input, GomoryHuOptions& options);
	~GomoryHu();
	void GetSolution(int* parents, GH_REAL* cut_costs);

private:
	int n;
	G* g;
	Tree* T;
	BufferList* buf_list;
	int depth;

	double T_selection_start, T_selection_time;
	int S_prev[2];
	int current_num[2]; // current_num[0]: free nodes, current_num[1]: fixed nodes
	int i_nonfixed; // set by PickTerminalsClosest()

	GomoryHu(int n, int m, Tree* T, BufferList* buf_list, int depth);
	void PickTerminalsFromScratch(int* s); // writes to s[0] and s[1]
	void PickTerminals(int* s); // writes to s[0] and s[1]
	void PickTerminals_HEAVIEST_HEAVIEST_EFFICIENT(int* s);
	void Solve();
	void SolveSmallerComponent(int num, int* node_list, int d, int terminal_prev);

	int* heavy_order;
	int* heavy_next;
};

GomoryHu::GomoryHu(TmpGraph<GH_REAL>* input, GomoryHuOptions& options)
	: n(input->n), depth(0), heavy_order(NULL)
{
	buf_list = new BufferList;
	T = new Tree(n);
	T->options = options;

	int i, e;

	S_prev[0] = S_prev[1] = -1;
	i_nonfixed = 0;

	g = new G(n, input->m, &buf_list->buf, &T->common_rbuf, buf_list->rbuf);
	for (e=0; e<input->m; e++)
	{
		g->add_edge(input->edges[2*e], input->edges[2*e+1], input->weights[e], input->weights[e]);
	}
	if (input->marks)
	{
		for (i=0; i<n; i++)
		{
			if (!input->marks[i])
			{
				g->nodes[i].gh_tree_node = T->NewNode();
				g->nodes[i].gh_fixed = 1;
			}
		}
	}
	Solve();
}

GomoryHu::GomoryHu(int _n, int m, Tree* _T, BufferList* _buf_list, int _depth)
	: n(_n), T(_T), buf_list(_buf_list), depth(_depth), heavy_order(NULL)
{
	S_prev[0] = S_prev[1] = -1;
	i_nonfixed = 0;
	g = new G(n, m, &buf_list->buf, &T->common_rbuf, buf_list->rbuf);
}

GomoryHu::~GomoryHu()
{
	delete g;
	if (depth == 0)
	{
		delete buf_list;
		delete T;
	}
	if (heavy_order) delete [] heavy_order;
}

void GomoryHu::GetSolution(int* parents, GH_REAL* cut_costs)
{
	int i;
	for (i=0; i<n; i++)
	{
		TreeNode* ti = g->nodes[i].gh_tree_node;
		parents[i] = (ti->parent) ? ti->parent->index : -1;
		cut_costs[i] = ti->cut_cost;
	}
}

const __int32_t MAX_DIST = (1 << 30);

void GomoryHu::PickTerminals_HEAVIEST_HEAVIEST_EFFICIENT(int* S)
{
	int i, k;
	G::arc* a;

	if (!heavy_order)
	{
		std::vector<int> V(n);
		GH_REAL* W = (GH_REAL*) T->common_rbuf.Alloc(n*sizeof(GH_REAL));
		k = 0;
		for (i=g->get_next_noncontracted(-1); i>=0; i=g->get_next_noncontracted(i))
		{
			if (g->nodes[i].gh_fixed) continue;
			W[k] = 0;
			FOR_ALL_ARCS(g->nodes+i, a) W[k] += GET_RCAP(a) + GET_RCAP(REV_ARC_(G, a));
			V[k ++] = i;
		}
		sort( V.begin(), V.begin()+k, [&](int i,int j) {return (W[i] > W[j]);} );
		heavy_order = new int[k];
		for (i=0; i<k; i++) heavy_order[i] = V[i];
		S[0] = heavy_order[0];
		S[1] = heavy_order[1];
		heavy_next = heavy_order + 2;
		return;
	}

	int d;
	for (d=0; d<2; d++)
	{
		if (S_prev[d] >= 0 && !g->nodes[S_prev[d]].gh_fixed) break;
	}
	if (d == 2 || g->nodes[S_prev[d]].is_contracted) { printf("Fatal error #1 in PickTerminals()\n"); exit(1); }
	S[d] = S_prev[d];
	while (g->nodes[*heavy_next].is_contracted || g->nodes[*heavy_next].gh_fixed) heavy_next ++;
	S[1-d] = *heavy_next;
	heavy_next ++;
}

void GomoryHu::PickTerminalsFromScratch(int* S)
{
	T_selection_start = get_time();

	int i, d;
	S[0] = S[1] = -1;
	int _s, _t;
	if (T->options.st_selection == GomoryHuOptions::RANDOM_RANDOM || T->options.st_selection == GomoryHuOptions::RANDOM_CLOSEST)
	{
		_s = rand() % current_num[0];
		if (T->options.st_selection == GomoryHuOptions::RANDOM_RANDOM)
		{
			_t = rand() % (current_num[0]-1); if (_t >= _s) _t ++;
			if (_s > _t) { i = _s; _s = _t; _t = i; }
			_t -= _s+1;
		}
		else _t = -1;

		for (i=g->get_next_noncontracted(-1); i>=0; i=g->get_next_noncontracted(i))
		{
			if (g->nodes[i].gh_fixed) continue;
			g->nodes[i].gh_dist[0] = g->nodes[i].gh_dist[1] = MAX_DIST;
			if      (S[0] < 0) { if (_s -- == 0) S[0] = i; }
			else if (S[1] < 0) { if (_t -- == 0) S[1] = i; }
		}
	}
	else
	{
		GH_REAL W_max[2] = { 0, 0 };
		for (i=g->get_next_noncontracted(-1); i>=0; i=g->get_next_noncontracted(i))
		{
			if (g->nodes[i].gh_fixed) continue;
			g->nodes[i].gh_dist[0] = g->nodes[i].gh_dist[1] = MAX_DIST;
			G::arc* a;
			GH_REAL W = 0;
			FOR_ALL_ARCS(g->nodes+i, a) W += GET_RCAP(a) + GET_RCAP(REV_ARC_(G, a));
			if (S[1] < 0 || W_max[1] < W)
			{
				S[1] = i; W_max[1] = W;
				if (S[0] < 0 || W_max[0] < W_max[1]) { int k = S[0]; S[0] = S[1]; S[1] = k; W = W_max[0]; W_max[0] = W_max[1]; W_max[1] = W; }
			}
		}
		if (T->options.st_selection == GomoryHuOptions::HEAVIEST_CLOSEST) S[1] = -1;
	}

	////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////

	for (d=0; d<2; d++)
	{
		G::node* i = g->nodes + S[d];
		G::node** queue = (G::node**) T->common_rbuf.Alloc(sizeof(G::node*));
		queue[0] = i;
		i->gh_dist[d] = 0;
		i->gh_smaller_component = 1;
		int from = 0, to = 1;
		G::node* i_next_level = NULL;
		__int32_t D = 1;
		while ( from < to )
		{
			i = queue[from ++];
			if (i == i_next_level)
			{
				i_next_level = NULL;
				if (D < MAX_DIST) D ++;
			}

			G::arc* a;
			FOR_ALL_ARCS(i, a)
			{
				G::node* j = a->head;
				if (j->gh_smaller_component) continue;
				if (!j->gh_fixed)
				{
					j->gh_dist[d] = D;
				}
				queue = (G::node**) T->common_rbuf.Realloc((to+1)*sizeof(G::node*));
				queue[to ++] = j;
				if (i_next_level == NULL) i_next_level = j;
				j->gh_smaller_component = 1;
			}
		}
		for (from=g->get_next_noncontracted(-1); from>=0; from=g->get_next_noncontracted(from)) g->nodes[from].gh_smaller_component = 0;

		if (d == 0 && S[1] < 0)
		{
			int D_min = 0, i;
			for (i=g->get_next_noncontracted(-1); i>=0; i=g->get_next_noncontracted(i))
			{
				if (g->nodes[i].gh_fixed || i == S[0]) continue;
				if (S[1] < 0 || D_min > g->nodes[i].gh_dist[0]) { S[1] = i; D_min = g->nodes[i].gh_dist[0]; }
			}
		}
	}

	T_selection_time = get_time() - T_selection_start;
}

void GomoryHu::PickTerminals(int* S)
{
	if (T->options.st_selection == GomoryHuOptions::HEAVIEST_HEAVIEST_EFFICIENT)
	{
		PickTerminals_HEAVIEST_HEAVIEST_EFFICIENT(S);
		return;
	}

	if (T->options.st_selection == GomoryHuOptions::CLOSEST_PAIR)
	{
		if (i_nonfixed >= n)
		{
			T->options.st_selection = GomoryHuOptions::HEAVIEST_HEAVIEST;
			PickTerminals(S);
			T->options.st_selection = GomoryHuOptions::CLOSEST_PAIR;
			return;
		}
		for ( ; i_nonfixed < n; i_nonfixed ++)
		{
			G::node* i = g->nodes + i_nonfixed;
			if (i->gh_fixed || i->is_contracted) continue;
			G::arc* a;
			FOR_ALL_ARCS(i, a)
			{
				G::node* j = a->head;
				if (!j->gh_fixed)
				{
					S[0] = (int)(i-g->nodes);
					S[1] = (int)(j-g->nodes);
					return;
				}
			}
			T->options.st_selection = GomoryHuOptions::HEAVIEST_HEAVIEST;
			PickTerminalsFromScratch(S);
			T->options.st_selection = GomoryHuOptions::CLOSEST_PAIR;
			return;
		}
	}

	if ( S_prev[0] < 0 || S_prev[1] < 0
		|| (T->options.st_freq > 0 && get_time() - T_selection_start > T_selection_time*T->options.st_freq) )
	{
		PickTerminalsFromScratch(S);
		return;
	}

	////////////////////////////////////////////////////////////////////////////////////
	int d;
	for (d=0; d<2; d++)
	{
		if (S_prev[d] >= 0 && !g->nodes[S_prev[d]].gh_fixed) break;
	}
	if (d == 2 || g->nodes[S_prev[d]].is_contracted) { printf("Fatal error #1 in PickTerminals()\n"); exit(1); }
	S[d] = S_prev[d];
	S[1-d] = -1;

	G::node* id = g->nodes + S_prev[d];
	G::node* i = g->nodes + S_prev[1-d];
	G::node** queue = (G::node**) T->common_rbuf.Alloc(sizeof(G::node*));
	queue[0] = i;
	i->gh_smaller_component = 1;
	int from = 0, to = 1;
	G::node* i_next_level = NULL;
	__int32_t D_min = MAX_DIST;
	while ( from < to )
	{
		i = queue[from ++];
		if (i == i_next_level)
		{
			if (S[1-d] >= 0) break;
			i_next_level = NULL;
		}

		G::arc* a;
		FOR_ALL_ARCS(i, a)
		{
			G::node* j = a->head;
			if (j->gh_smaller_component) continue;
			if (!j->gh_fixed && j!=id)
			{
				if (D_min > j->gh_dist[d])
				{
					D_min = j->gh_dist[d];
					S[1-d] = (int) (j-g->nodes);
				}
			}
			queue = (G::node**) T->common_rbuf.Realloc((to+1)*sizeof(G::node*));
			queue[to ++] = j;
			if (i_next_level == NULL) i_next_level = j;
			j->gh_smaller_component = 1;
		}
	}
	for (from=0; from<to; from++) queue[from]->gh_smaller_component = 0;

	if (S[1-d] < 0)
	{
		PickTerminalsFromScratch(S);
	}
}


void GomoryHu::Solve()
{
	assert(T->options.threshold > 0);
	if (n == 0) return;

	int i, s, t;
	G::arc* a;

	TreeNode* special_first = NULL;
	int special_num = 0;

	current_num[0] = current_num[1] = 0;
	GH_REAL threshold_orig = T->options.threshold;

	for (i=0; i<n; i++)
	{
		current_num[g->nodes[i].gh_fixed] ++;
	}

	if (current_num[0] == 0)
	{
		for (i=0; i<n; i++)
		{
			g->nodes[i].gh_tree_node->index = i;
			if (i>0) g->nodes[i].gh_tree_node->parent = g->nodes[0].gh_tree_node;
		}
		return;
	}

	bool first_maxflow = true;
	g->init_graph();

	while (current_num[0] > 1)
	{
		int S[2];
		PickTerminals(S);
		s = S[0]; t = S[1];
		assert( s != t && s >= 0 && t >= 0 );
		assert(!g->nodes[s].gh_fixed && !g->nodes[s].is_contracted);
		assert(!g->nodes[t].gh_fixed && !g->nodes[t].is_contracted);

		double T_start = get_time();
		if (first_maxflow)
		{
			UPDATE_EXCESS_FIRST(g, s, T->options.threshold, 0);
			UPDATE_EXCESS_FIRST(g, t, 0, T->options.threshold);
			MAXFLOW_FIRST(g);
			first_maxflow = false;
		}
		else
		{
			UPDATE_EXCESS_INC(g, s, T->options.threshold, 0);
			UPDATE_EXCESS_INC(g, t, 0, T->options.threshold);
			MAXFLOW_INC(g);
		}
		maxflow_time += get_time() - T_start;
		maxflow_num ++;
		maxflow_total_node_num += g->node_num_noncontracted;
		maxflow_total_edge_num += g->edge_num;

		int num;
		int* list;
#ifdef KEEP_STLISTS
		int d = g->get_smaller_component(num, list, &buf_list->rbuf[MAXFLOW_RBUF_NUM]);
#else
		int d = g->get_smaller_balanced_component(num, list, &buf_list->rbuf[MAXFLOW_RBUF_NUM]);
#endif

		int s_segm = 1-d, t_segm = 1-d;
		for (i=0; i<num; i++)
		{
			if      (list[i] == s) s_segm = d;
			else if (list[i] == t) t_segm = d;
		}

		UPDATE_EXCESS_INC(g, s, -T->options.threshold, 0);
		UPDATE_EXCESS_INC(g, t, 0, -T->options.threshold);

		if (s_segm == 0 && t_segm == 1)
		{
			if (T->options.verbose >= 2 && depth == 0)
			{
				std::cout << "(" << s << "," << t << "): contracting";
				for (i=0; i<num; i++) std::cout << " " << list[i];
				std::cout << "\n";
			}
			S_prev[1-d] = S[1-d];
			S_prev[d] = list[0];

			SolveSmallerComponent(num, list, d, S[d]);
			if (special_num < T->options.decrease_threshold_condition*(special_num + current_num[0]))
			{
				GH_REAL t_orig = T->options.threshold;
				T->options.threshold = (GH_REAL)(T->options.decrease_threshold_factor*t_orig);
				if (T->options.threshold <= 0) T->options.threshold = t_orig;
			}
		}
		else
		{
			if (T->options.verbose >= 2 && depth == 0)
			{
				std::cout << "(" << s << "," << t << "): merging\n";
			}
			S_prev[0] = S_prev[1] = s;
			int list[2];
			list[0] = s;
			list[1] = t;
			g->contract(2, list);
			TreeNode* r = T->NewNode();
			g->nodes[t].gh_tree_node = r;
			g->nodes[t].gh_fixed = 1;
			r->index = s;
			r->sibling = special_first;
			r->parent = r;
			r->cut_cost = -T->options.threshold;
			special_first = r;
			current_num[0] --;
			special_num ++;
			merge_num ++;
		}
	}

	s = -1;
	TreeNode* r = T->NewNode();
	r->cut_cost = 0;
	for (i=g->get_next_noncontracted(-1); i>=0; i=g->get_next_noncontracted(i))
	{
		if (!g->nodes[i].gh_fixed)
		{
			assert(s < 0);
			s = i;
			continue;
		}
		TreeNode* ti = g->nodes[i].gh_tree_node;
		ti->parent = r;
		ti->sibling = r->first_child;
		r->first_child = ti;
		ti->cut_cost = 0;
		FOR_ALL_ARCS(g->nodes+i, a)
		{
			GH_REAL w = (GET_RCAP(a) + GET_RCAP(REV_ARC_(G, a))) / 2;
			ti->cut_cost += w;
		}
	}
	assert( s >= 0 );
	g->nodes[s].gh_tree_node = r;
	g->nodes[s].gh_fixed = 1;

	while (special_first)
	{
		TreeNode* i = special_first;
		special_first = i->sibling;
		TreeNode* p = g->nodes[i->index].gh_tree_node;
		i->parent = p;
		i->sibling = p->first_child;
		p->first_child = i;
	}

	if (depth == 0)
	{
		for (i=0; i<n; i++)
		{
			g->nodes[i].gh_tree_node->index = i;
		}
	}

	T->options.threshold = threshold_orig;
}



void GomoryHu::SolveSmallerComponent(int num, int* node_list, int d, int terminal_prev)
{
	int k;

	BufferList* buf_list_new = buf_list->GetNext();
	GomoryHu S(num+1, g->edge_num, T, buf_list_new, depth+1);

	int id_current = 1;
	for (k=0; k<num; k++)
	{
		G::node* i = g->nodes + node_list[k];
		i->gh_smaller_component = 1;
		i->gh_index = id_current ++;
		current_num[i->gh_fixed] --;
	}

	//GH_REAL cut_value = 0;
	for (k=0; k<num; k++)
	{
		G::node* i = g->nodes + node_list[k];
		S.g->nodes[i->gh_index].gh_tree_node = i->gh_tree_node;
		S.g->nodes[i->gh_index].gh_fixed = i->gh_fixed;

		G::arc* a;
		GH_REAL r0 = 0, r0_rev = 0;
		FOR_ALL_ARCS(i, a)
		{
			G::node* j = a->head;
			if (j->gh_smaller_component)
			{
				if (i < j)
				{
					GH_REAL w = (GET_RCAP(a) + GET_RCAP(REV_ARC_(G, a))) / 2;
					S.g->add_edge(i->gh_index, j->gh_index, w, w);
				}
			}
			else
			{
				r0 += GET_RCAP(a);
				r0_rev += GET_RCAP(REV_ARC_(G, a));
			}
		}
		GH_REAL w = (r0 + r0_rev) / 2;
		if (w > 0)
		{
			S.g->add_edge(0, i->gh_index, w, w);
			//cut_value += w;
		}
	}
	TreeNode root;
	memset(&root, 0, sizeof(TreeNode));
	S.g->nodes[0].gh_tree_node = &root;
	S.g->nodes[0].gh_fixed = 1;
	S.Solve();
	T->ChangeRoot(&root);
	for (k=0; k<num; k++)
	{
		G::node* i = g->nodes + node_list[k];
		TreeNode* ti = S.g->nodes[i->gh_index].gh_tree_node;
		ti->index = node_list[k];
		if (i->gh_fixed) ti->index += n;
		else
		{
			i->gh_tree_node = ti;
			i->gh_fixed = 1;
		}
	}

	TreeNode* r;
	TreeNode* r_sibling;
	for (r=root.first_child; r; r=r_sibling)
	{
		r_sibling = r->sibling;
		r->parent = r->sibling = NULL;
		int num = 0;
		TreeNode* i = r;
		int ir = node_list[num ++] = i->index % n;
		if ( r->first_child )
		{
			while ( 1 )
			{
				if (i->index < n && i->first_child)
				{
					i = i->first_child;
					continue;
				}
				while ( 1 )
				{
					node_list[num ++] = i->index % n;
					if (i->sibling) break;
					i = i->parent;
					if (i == r) break;
				}
				if (i == r) break;
				i = i->sibling;
			}
		}
		g->contract(num, node_list);
		g->nodes[node_list[0]].gh_smaller_component = 0;
		current_num[1] ++;
	}
	buf_list_new->buf.Reset();
}



////////////////////////////////////////////////////////////////////////////////



void _SolveGomoryHu(TmpGraph<GH_REAL>* input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options)
{
	GomoryHu S(input, options);
	S.GetSolution(parents, cut_costs);
}

