#ifndef __GOMORY_HU__
#define __GOMORY_HU__

#include "timer.h"
#include "tmp_graph.h"

typedef int GH_REAL;

#define REDUCTION_REORDER_EDGES        0x0001
#define REDUCTION_REORDER_NODES        0x0002
#define REDUCTION_CC                   0x0004
#define REDUCTION_MERGE_PARALLEL_EDGES 0x0008

struct GomoryHuOptions
{
	GomoryHuOptions()
		: threshold(0),
		  method(METHOD_NAIVE),
		  reduction_method(REDUCTION_REORDER_NODES | REDUCTION_CC),
		  st_selection(HEAVIEST_HEAVIEST),
		  st_freq(0),
		  verbose(1),
		  decrease_threshold_condition(0), decrease_threshold_factor(1)
	{
	}
	void Print();

	typedef enum
	{
		METHOD_NAIVE = 0,
		METHOD_D = 1, // experimental
		METHOD_T = 2, // experimental
		METHOD_A = 3  // experimental
	} Method;

	GH_REAL threshold; // 0 converted to (1 + max weighted node degree)

	Method method;
	int stage_num;
	double* D_array; // if method == METHOD_N: call solver 'stage_num+1' times
				     // at stage 'i' use ngood_threshold = N_array[i], at the last stage use ngood_threshold = 0
	GH_REAL* T_array; // if method == METHOD_T: call solver 'stage_num+1' times with the following thresholds:
	               // T_array[0], T_array[1], ..., T_array[stage_num-1], threshold.
	               // this sequence must be nondecreasing.
	double* A_array; // if method == METHOD_A: use METHOD_T with T_array[i] = max(1, (REAL)(A_array*threshold))

	typedef enum
	{
		RANDOM_RANDOM = 0,
		RANDOM_CLOSEST = 1,
		HEAVIEST_HEAVIEST = 2,
		HEAVIEST_CLOSEST = 3,
		CLOSEST_PAIR = 4,
		HEAVIEST_HEAVIEST_EFFICIENT = 5 // st_freq is then ignored
	} stSelection;
	stSelection st_selection;

	double st_freq; // how frequently choose new s-t pair. If 0: never update the pair. Otherwise: larger value means less frequently

	int reduction_method;

	int verbose;


	// should not be set directly
	double decrease_threshold_condition; // \in [0,1]
	double decrease_threshold_factor; // \in (0,1]
};


// Note, the call will change 'input', if options.reduction_method != 0
void SolveGomoryHu(TmpGraph<GH_REAL>* input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options);



// useful util function
inline void ComputePostorder(int n, int* parents, int* postorder)
{
	int i;
	int* child = new int[2*n];
	int* sibling = child + n;

	for (i=0; i<n; i++) child[i] = -1;
	int root = -1;
	for (i=0; i<n; i++)
	{
		int p = parents[i];
		if (p >= 0)
		{
			sibling[i] = child[p];
			child[p] = i;
		}
		else root = i;
	}

	i = root;
	int k = 0;
	while ( 1 )
	{
		if (child[i] >= 0)
		{
			i = child[i];
			continue;
		}
		while ( 1 )
		{
			postorder[k ++] = i;
			if (sibling[i] >= 0) break;
			i = parents[i];
			if (i == root) { postorder[k] = i; break; }
		}
		if (i == root) break;
		i = sibling[i];
	}

	delete [] child;
}

#endif
