#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include "graph_with_tree.h"
#include "gomory_hu_reductions.h"

void _SolveGomoryHu(TmpGraph<GH_REAL>* g, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options);

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

class GomoryHuNested : GraphWithTree
{
public:
	GomoryHuNested(int n, int* parents, GH_REAL* cut_costs);
	~GomoryHuNested();

	void AddEdges(TmpGraph<GH_REAL>* input);
	void Solve(int* parents, GH_REAL* cut_costs, GomoryHuOptions& options);

	GraphWithTreeAndLCA* g;
	GraphWithTree* h;
	int N;
	Node** components;
};


GomoryHuNested::GomoryHuNested(int n, int* parents, GH_REAL* cut_costs)
	: GraphWithTree(n), h(NULL)
{
	int _i, k;
	Node* i;
	for (_i=0; _i<n; _i++)
	{
		if (parents[_i] >= 0)
		{
			nodes[_i].is_cut = (cut_costs[_i] >= 0) ? true : false;
			AddTreeEdge(_i, parents[_i]);
		}
		else
		{
			nodes[_i].is_cut = true;
			AddRoot(_i);
		}
	}

	N = 0;
	for (k=n-1; k>=0; k--)
	{
		_i = postorder[k]; i = nodes + _i;
		if (i->is_cut) i->index = N ++;
		else           i->index = i->tree_parent->index;
	}

	g = new GraphWithTreeAndLCA(N);
	for (_i=0; _i<n; _i++)
	{
		i = nodes + _i;
		if (i->is_cut)
		{
			if (i->tree_parent) g->AddTreeEdge(i->index, i->tree_parent->index);
			else                g->AddRoot(i->index);
		}
	}
	g->SetLCA();

	components = new Node*[N];
	memset(components, 0, N*sizeof(Node*));
	for (k=0; k<n; k++)
	{
		_i = postorder[k]; i = nodes + _i;
		i->next = components[i->index];
		components[i->index] = i;
	}
}

GomoryHuNested::~GomoryHuNested()
{
	delete [] components;
	delete g;
	if (h) delete h;
}

void GomoryHuNested::AddEdges(TmpGraph<GH_REAL>* input)
{
	int e;
	for (e=0; e<input->m; e++) AddEdge(input->edges[2*e], input->edges[2*e+1], input->weights[e]);

	int c;
	for (c=0; c<N; c++)
	{
		Node* i = components[c];
		int index = i->index;
		g->StartNode(index);
		for ( ; i; i=i->next)
		{
			Arc* a;
			for (a=i->first; a; a=a->next)
			{
				if (a->head->index > index) g->AddEdge(a->head->index, a->w);
			}
		}
		g->FinishNode();
	}
	h = g->TransformGraph();
}

void ChangeRoot(int n, int* parents, GH_REAL* costs, int new_root)
{
	int i, D=0;
	for (i=new_root; parents[i]>=0; i=parents[i]) D ++;
	int* path = new int[++ D];
	int d = 0;
	for (i=new_root; parents[i]>=0; i=parents[i]) path[d ++] = i;
	path[d] = i;
	for (d=D-1; d>0; d--)
	{
		costs[path[d]] = costs[path[d-1]];
		parents[path[d]] = path[d-1];
	}
	parents[new_root] = -1;
	costs[new_root] = 0;
	delete [] path;
}

void GomoryHuNested::Solve(int* PARENTS, GH_REAL* CUT_COSTS, GomoryHuOptions& options)
{
	double T_start = get_time();

	int c;
	Node* i;
	Arc* a;

	int index0_max[2] = { 0, 0 }, index_max[2] = { 0, 0 };

	for (c=0; c<N; c++)
	{
		int index = 0;
		for (i=components[c]; i; i=i->next) i->index = -1-(index ++);
		int index0 = index;
		if (index0 <= 1)
		{
			for (i=components[c]; i; i=i->next) i->index = c;
			continue;
		}
		if (g->nodes[c].tree_parent) g->nodes[c].tree_parent->index = index ++;
		for (i=g->nodes[c].tree_child; i; i=i->tree_sibling) i->index = index ++;

		if (index_max[1] < index) { index_max[1] = index; index0_max[1] = index0; }
		if (index_max[0] < index_max[1]) { int a = index_max[0]; index_max[0] = index_max[1]; index_max[1] = a; a = index0_max[0]; index0_max[0] = index0_max[1]; index0_max[1] = a; }

		Node* I = g->nodes + c;
		Node* J;
		GraphWithTree* gc = new GraphWithTree(index);
		int M = 0;
		for (i=components[c]; i; i=i->next)
		{
			gc->StartNode(-1-i->index);
			for (a=i->first; a; a=a->next)
			{
				Node* j = a->head;
				if (j->index < 0)
				{
					if (i < j)
					{
						if (gc->AddEdge(-1-j->index, a->w)) M ++;
					}
				}
				else
				{
					J = g->nodes + j->index;
					Node* II = I;
					g->getPenultimateNodes(II, J);
					int Jindex = (J->tree_parent == I) ? J->index : index0;
					if (gc->AddEdge(Jindex, a->w)) M ++;
				}
			}
			gc->FinishNode();
		}

		for (J=I->tree_child; J; J=J->tree_sibling)
		{
			Node* Jh = h->nodes + (int)(J-g->nodes);
			for (a=Jh->first; a; a=a->next)
			{
				Node* K = g->nodes + (int)(a->head - h->nodes);
				if (K > J) { gc->AddEdge(J->index, K->index, a->w); M ++; }
			}

			if (Jh->w_down > 0) { gc->AddEdge(J->index, index0, Jh->w_down); M ++; }
		}

		TmpGraph<GH_REAL>* input = new TmpGraph<GH_REAL>(index, M);
		for (i=gc->nodes; i<gc->nodes+index; i++)
		{
			for (a=i->first; a; a=a->next)
			{
				Node* j = a->head;
				if (j > i)
				{
					input->AddEdge((int)(i-gc->nodes), (int)(j-gc->nodes), a->w);
				}
			}
		}
		if (input->m != M) { printf("Fatal error in GomoryHuNested::Solve()\n"); exit(1); }
		delete gc;

		int* parents = new int[index];
		GH_REAL* cut_costs = new GH_REAL[index];
		int k;
		for (k=index0; k<index; k++) input->UnsetMark(k);
		_SolveGomoryHu(input, parents, cut_costs, options);
		delete input;

		Node* p = components[c]->tree_parent;
		ChangeRoot(index, parents, cut_costs, p ? index0 : 0);

		int* map = (new int[1+index0+1])+1;
		map[-1] = -1;
		for (i=components[c]; i; i=i->next) map[-1 - i->index] = (int)(i-nodes);
		if (p) map[index0] = (int)(p - nodes);
		for (i=components[c]; i; i=i->next)
		{
			int _i = (int)(i-nodes);
			PARENTS[_i] = map[parents[-1 - i->index]];
			nodes[_i].tree_parent = nodes + PARENTS[_i];
			CUT_COSTS[_i] = cut_costs[-1 - i->index];
		}
		for (i=g->nodes[c].tree_child; i; i=i->tree_sibling)
		{
			int _i = (int)(components[(int)(i - g->nodes)]-nodes);
			PARENTS[_i] = map[parents[i->index]];
			nodes[_i].tree_parent = nodes + PARENTS[_i];
		}
		delete [] (map-1);

		delete [] parents;
		delete [] cut_costs;

		for (i=components[c]; i; i=i->next) i->index = c;
	}

	if (options.verbose >= 1)
	{
		printf("component sizes: %d => (%d %d), (%d %d), ... [%f secs]\n", node_num, index_max[0], index0_max[0], index_max[1], index0_max[1], get_time()-T_start);
	}
}

extern int maxflow_num;
extern int merge_num;
extern double maxflow_time;
extern double maxflow_total_node_num;
extern double maxflow_total_edge_num;


int ComputeTreeDiameter(int n, int* parents)
{
	int k, i;
	int* D = new int[3*n];
	int* E = D + n;
	int* postorder = D + 2*n;
	memset(D, 0, n*sizeof(int));
	ComputePostorder(n, parents, postorder);
	
	int D_max = 0;
	E[postorder[n-1]] = 0;
	for (k=n-2; k>=0; k--)
	{
		i = postorder[k];
		E[i] = 1 + E[parents[i]];
	}
	for (k=0; k<n-1; k++)
	{
		i = postorder[k];
		if (D_max < D[i] + E[i]) D_max = D[i] + E[i];
		int p = parents[i];
		int Dp = D[i] + 1;
		if (D_max < D[p] + Dp) D_max = D[p] + Dp;
		if (D[p] < Dp) D[p] = Dp;
	}
	delete [] D;
	return D_max;
}

void PreprocessOptions(TmpGraph<GH_REAL> *input, GomoryHuOptions& options)
{
	int n = input->n;

	int k;
	GH_REAL threshold_max = 0;
	GH_REAL* W = new GH_REAL[n];
	memset(W, 0, n*sizeof(GH_REAL));
	int i, e;
	for (e=0; e<2*input->m; e++) W[input->edges[e]] += input->weights[e/2];
	for (i=0; i<n; i++)
	{
		if (threshold_max < W[i]) threshold_max = W[i];
	}
	threshold_max += 1;
	delete [] W;

	if (options.threshold == 0 || options.threshold > threshold_max) options.threshold = threshold_max;
	threshold_max = options.threshold;

	if (options.method == GomoryHuOptions::METHOD_A)
	{
		options.T_array = new GH_REAL[options.stage_num];
		for (k=options.stage_num-1; k>=0; k--)
		{
			options.T_array[k] = (GH_REAL)(threshold_max*options.A_array[k]);
			if (options.T_array[k] == 0) options.T_array[k] = (k == options.stage_num-1) ? threshold_max : options.T_array[k+1];
		}
		options.method = GomoryHuOptions::METHOD_T;
	}

	if (options.method == GomoryHuOptions::METHOD_T) 
	{
		options.threshold = options.T_array[options.stage_num-1];
	}
}

void PostprocessOptions(TmpGraph<GH_REAL> *input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options_orig, GomoryHuOptions& options)
{
	if (options_orig.method == GomoryHuOptions::METHOD_A)
	{
		delete [] options.T_array;
	}
}

void SolveGomoryHu(TmpGraph<GH_REAL> *input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options)
{
	double T_start = get_time();
	maxflow_num = merge_num = 0;
	maxflow_time = maxflow_total_node_num = maxflow_total_edge_num = 0;
	int i, k, n = input->n, m = input->m;

	GomoryHuOptions options_orig = options;
	PreprocessOptions(input, options);

	if (options.verbose >= 1) options.Print();

	GomoryHuReductions R(input, parents, cut_costs, options.reduction_method);
	TmpGraph<int>* _input;
	int* _parents;
	GH_REAL* _cut_costs;
	while ((_input=R.GetNextSubproblem(_parents, _cut_costs)))
	{
		if (options.method == GomoryHuOptions::METHOD_NAIVE)
		{
			_SolveGomoryHu(_input, _parents, _cut_costs, options);
		}
		else if (options.method == GomoryHuOptions::METHOD_T)
		{
			options.threshold = options.T_array[0];
			_SolveGomoryHu(_input, _parents, _cut_costs, options);
			if (options.verbose >= 1 && options.stage_num > 1)
			{
				printf("first call: %f secs\n", get_time() - T_start);
			}
			for (k=1; k<options.stage_num; k++)
			{
				GomoryHuNested G(n, _parents, _cut_costs);
				G.AddEdges(_input);
				options.threshold = options.T_array[k];
				G.Solve(_parents, _cut_costs, options);
			}
		}
		else if (options.method == GomoryHuOptions::METHOD_D)
		{
			options.decrease_threshold_condition = options.D_array[0];
			options.decrease_threshold_factor = options.D_array[1];
			_SolveGomoryHu(_input, _parents, _cut_costs, options);
			if (options.verbose >= 1)
			{
				printf("first call: %f secs\n", get_time() - T_start);
			}
			for (k=1; k<=options.stage_num; k++)
			{
				GomoryHuNested G(n, _parents, _cut_costs);
				G.AddEdges(_input);
				if (k == options.stage_num)
				{
					options.decrease_threshold_condition = 0;
					options.decrease_threshold_factor = 1;
				}
				else
				{
					options.decrease_threshold_condition = options.D_array[2*k];
					options.decrease_threshold_factor = options.D_array[2*k+1];
				}
				G.Solve(_parents, _cut_costs, options);
			}
		}
		else exit(1);
	}

	PostprocessOptions(input, parents, cut_costs, options_orig, options);
	for (i=0; i<n; i++)
	{
		if (cut_costs[i] < 0) cut_costs[i] = options.threshold;
	}

	if (options.verbose >= 1)
	{
		GH_REAL cut_min = 0, cut_max = 0;
		double cut_sum = 0;
		int checksum = 0;
		bool flag = false;
		for (i=0; i<n; i++)
		{
			if (parents[i] < 0) continue;
			if (!flag) { cut_min = cut_max = cut_costs[i]; flag = true; }
			else
			{
				if (cut_min > cut_costs[i]) cut_min = cut_costs[i];
				if (cut_max < cut_costs[i]) cut_max = cut_costs[i];
			}
			cut_sum += cut_costs[i];
			checksum = ((int)(checksum + cut_costs[i])) % 1234;
		}
		std::cout << "cut values: (min,max,average)=(" << cut_min << "," << cut_max << "," << cut_sum/(n-1) << "), checksum=" << checksum << "\n";
		std::cout << "graph size: (n,m) = (" << n << "," << m << "), ";
		std::cout << "tree diameter: " << ComputeTreeDiameter(n, parents) << "\n";
		std::cout << maxflow_num << " maxflows, total graph size = (n*" << ((double)maxflow_total_node_num) / n << ", m*"
			      << ((double)maxflow_total_edge_num) / m << ")\n";
		std::cout << "maxflow time: " << maxflow_time << " secs\n";
		std::cout << "total   time: " << get_time() - T_start << " secs\n";
	}

	GH_REAL threshold = options.threshold;
	options = options_orig;
	options.threshold = threshold;
}

