
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

#define GRAPH_BASE_FORWARD_STAR
#ifndef DO_NOT_KEEP_SATURATION_FLAG
#define KEEP_SATURATION_FLAG
#endif

#ifdef KEEP_SATURATION_FLAG
	#define ARC_WITH_BIT_ONE(a)  ( (arc*)       (   ((std::uintptr_t)(a))   |           (1)             ) )
	#define ARC_WITH_BIT_ZERO(a) ( (arc*)       (   ((std::uintptr_t)(a))   &   (~((std::uintptr_t)1))  ) )
	#define ARC_WITH_RCAP(a, c) (((c) > 0) ? ARC_WITH_BIT_ONE(a) : ARC_WITH_BIT_ZERO(a))
	#define REV_ARC(a) ARC_WITH_BIT_ZERO((a)->_rev)
	#define REV_ARC_(G, a) ( (G::arc*)       (   ((std::uintptr_t)((a)->_rev))   &   (~((std::uintptr_t)1))  ) )
	#define IS_NONSATURATED(a) ((a)->r_cap > 0)
	#define IS_REV_NONSATURATED(a) (((std::uintptr_t)((a)->_rev)) & 1)
	#define GET_RCAP(a) ((a)->r_cap)
	#define SET_RCAP(a, c) { (a)->r_cap = c; (a)->_rev = ARC_WITH_RCAP((a)->_rev, c); }
	#define PUSH_FLOW(a, flow) { (a)->r_cap -= flow; REV_ARC(a)->r_cap += flow; (a)->_rev = ARC_WITH_RCAP((a)->_rev, REV_ARC(a)->r_cap); REV_ARC(a)->_rev = ARC_WITH_RCAP(REV_ARC(a)->_rev, (a)->r_cap); }
#else
	#define REV_ARC(a) (a->_rev)
	#define REV_ARC_(G, a) (a->_rev)
	#define ARC_WITH_BIT_ONE(a)  (a)
	#define ARC_WITH_BIT_ZERO(a) (a)
	#define ARC_WITH_RCAP(a, c) (a)

	#define IS_NONSATURATED(a) ((a)->r_cap > 0)
	#define IS_REV_NONSATURATED(a) (REV_ARC(a)->r_cap > 0)
	#define GET_RCAP(a) ((a)->r_cap)
	#define SET_RCAP(a, c) { (a)->r_cap = c; }
	#define PUSH_FLOW(a, flow) { (a)->r_cap -= flow; REV_ARC(a)->r_cap += flow; }
#endif



#ifdef GRAPH_CONTRACTION
	#define FOR_ALL_NODES(i) for (i=get_next_noncontracted(-1); i>=0; i=get_next_noncontracted(i))
	#define IS_VALID_ARC(i, a) ((a) != (i)->last)
#else
	#define FOR_ALL_NODES(i) for (i=0; i<node_num; i++)
	#define IS_VALID_ARC(i, a) ((a) != (i+1)->first)
#endif
#define FIRST_ARC(i) ((i)->first)
#define NEXT_ARC(a) ((a)+1)
#define FOR_ALL_ARCS(i, a) for (a=(i)->first; IS_VALID_ARC(i, a); a = (a) + 1)
#define FOR_REMAINING_ARCS(i, a) for ( ; IS_VALID_ARC(i, a); a = (a) + 1)


#ifdef KEEP_STLISTS
	#define STLISTS_RBUF_NUM 2
	// SET_EXCESS(i, c) will be defined a bit later
#else
	#define STLISTS_RBUF_NUM 0
	#define SET_EXCESS(i, c) { (i)->excess = c; }
#endif
#define UPDATE_EXCESS(i, delta) SET_EXCESS(i, (i)->excess + (delta))

////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

struct GraphBase
{
	typedef int node_id;
	typedef int arc_id;

#ifndef GRAPH_BASE_REAL
#define GRAPH_BASE_REAL int
#endif

	typedef GRAPH_BASE_REAL tcaptype;
	typedef GRAPH_BASE_REAL captype;
	typedef GRAPH_BASE_REAL flowtype;

	// rbuf is of size 	STLISTS_RBUF_NUM. common_rbuf can be used outside. rbuf[0] and rbuf[1] cannot.
	GraphBase(int _node_num, int edge_num_max, Buffer* buf, ReusableBuffer* common_rbuf, ReusableBuffer* rbuf);
	~GraphBase();


	// Adds a bidirectional edge between 'i' and 'j' with the weights 'cap' and 'rev_cap'.
	// IMPORTANT: see note about the constructor 
	void add_edge(node_id i, node_id j, captype cap, captype rev_cap);
	void init_graph();

	void init_graph(TmpGraph<captype>* g);

	struct arc;
	struct node
	{
		union
		{
			arc*		first;
			int			tmp_degree; // used in init_graph()
		};
#ifdef GRAPH_CONTRACTION
		arc*		last;
		union
		{
			arc*		a_contraction;
			int			next_noncontracted;
		};
		unsigned int is_contracted : 1;
#endif
#ifdef KEEP_STLISTS
		unsigned int is_in_source_list : 1;
		unsigned int is_in_sink_list : 1;
#endif
		unsigned int segment : 2; // 0: with source, 1: with sink, 2: disconnected from both

		tcaptype	excess;

#ifdef NODE_FIELDS
		NODE_FIELDS
#endif
#ifdef EXTRA_NODE_FIELDS
		EXTRA_NODE_FIELDS
#endif
	};

	struct arc
	{
		node*		head;
		arc*		_rev;
		captype		r_cap;

#ifdef EXTRA_ARC_FIELDS
		EXTRA_ARC_FIELDS
#endif
	};

	node				*nodes;
	arc					*arcs;
	int					node_num, edge_num;
#ifdef GRAPH_CONTRACTION
	int					node_num_noncontracted;
#else
	#define node_num_noncontracted node_num
#endif

#ifdef GRAPH_CONTRACTION
	void contract(int num, node_id* list);
	node_id get_next_noncontracted(node_id prev)
	{
		node_id i = prev + 1;
		if (i == node_num) return -1;
		if (!nodes[i].is_contracted) return i;
		node_id j = i;
		while (j < node_num && nodes[j].is_contracted)
		{
			if (nodes[j].next_noncontracted) j = nodes[j].next_noncontracted;
			else j ++;
		}
		nodes[i].next_noncontracted = j;
		return (j == node_num) ? -1 : j;
	}
#endif

	int get_smaller_component(int& num, int*& list, ReusableBuffer* rbuf);
	int get_smaller_balanced_component(int& num, int*& list, ReusableBuffer* rbuf);
	void set_segment_flags();

	void print();

	void check_cut_optimality()
	{
		int _i;
		FOR_ALL_NODES(_i)
		{
			node* i = nodes + _i;
			arc* a;
			if (i->segment == 0 && i->excess < 0) { printf("non-opimal cut! (#1)\n"); exit(1); }
			if (i->segment != 0 && i->excess > 0) { printf("non-opimal cut! (#2)\n"); exit(1); }
			if (i->segment != 0) continue;
			FOR_ALL_ARCS(i, a)
			{
				if (IS_NONSATURATED(a) && a->head->segment != 0) { printf("non-opimal cut! (#3)\n"); exit(1); }
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef KEEP_STLISTS
	ReusableBuffer*		srbuf[2];
	node**				slist[2];
	int					snum[2];

#define ADD_TO_SLIST(s, i) { slist[s] = (node**) srbuf[s]->Realloc((snum[s] + 1)*sizeof(node*)); slist[s][snum[s] ++] = i; }
#define SET_EXCESS(i, c) \
	{ \
		(i)->excess = c; \
		if      ((i)->excess > 0) { if (!(i)->is_in_source_list) { (i)->is_in_source_list = 1; ADD_TO_SLIST(0, i); } } \
		else if ((i)->excess < 0) { if (!(i)->is_in_sink_list)   { (i)->is_in_sink_list   = 1; ADD_TO_SLIST(1, i); } } \
	}
#endif

	Buffer*				buf;
	ReusableBuffer*		common_rbuf;

private:
	struct TmpCapTypePair { TmpCapTypePair(captype _a, captype _b) : a(_a), b(_b) {} captype a, b; };
	TmpGraph<TmpCapTypePair>* tmp_graph;
};

// sizeof(arc) must be even. If not, the statement below will not compile.
extern char dummy_array[2*(  (1+sizeof(GraphBase::arc))&1  )-1];

////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

inline GraphBase::GraphBase(int _node_num, int edge_num_max, Buffer* _buf, ReusableBuffer* _common_rbuf, ReusableBuffer* _rbuf) 
	: node_num(_node_num), edge_num(0), buf(_buf), common_rbuf(_common_rbuf)
#ifdef GRAPH_CONTRACTION
	, node_num_noncontracted(_node_num)
#endif
{
	nodes = (node*) buf->AllocAligned((node_num+1)*sizeof(node), 64);
	memset(nodes, 0, (node_num+1)*sizeof(node));
	arcs = (arc*) buf->AllocAligned(2*edge_num_max*sizeof(arc), 64);
	tmp_graph = new TmpGraph<TmpCapTypePair>(node_num, edge_num_max);

#ifdef KEEP_STLISTS
	srbuf[0] = &_rbuf[0];
	srbuf[1] = &_rbuf[1];
	snum[0] = snum[1] = 0;
#endif
}

inline GraphBase::~GraphBase()
{
	if (tmp_graph) delete tmp_graph;
}

inline void GraphBase::add_edge(node_id i, node_id j, captype cap, captype rev_cap)
{
	tmp_graph->AddEdge(i, j, TmpCapTypePair(cap, rev_cap));
	edge_num ++;
}

inline void GraphBase::init_graph()
{
	if (!tmp_graph) return;

	int e;
	node* i;
	node* j;
	//	TmpGraph<TmpCapTypePair>::Edge *e;
	for (e=0; e<2*tmp_graph->m; e++) nodes[tmp_graph->edges[e]].tmp_degree ++;
	arc* a = arcs;
	for (i=nodes; i<=nodes+node_num; i++)
	{
		int degree = i->tmp_degree;
		i->first = a;
		a += degree;
	}
	for (e=0; e<tmp_graph->m; e++)
	{
		i = nodes + tmp_graph->edges[2*e];
		j = nodes + tmp_graph->edges[2*e+1];
		a = i->first ++;
		arc* a_rev = j->first ++;
		a->r_cap = tmp_graph->weights[e].a;
		a_rev->r_cap = tmp_graph->weights[e].b;
		a->_rev     = ARC_WITH_RCAP(a_rev, a_rev->r_cap);
		a_rev->_rev = ARC_WITH_RCAP(a, a->r_cap);

		a->head = j;
		a_rev->head = i;
	}
	a = arcs;
	for (i=nodes; i<nodes+node_num; i++)
	{
		arc* tmp = i->first;
		i->first = a;
#ifdef GRAPH_CONTRACTION
		if (i > nodes) (i-1)->last = i->first;
#endif
		a = tmp;
	}
#ifdef GRAPH_CONTRACTION
	(i-1)->last = i->first;
#endif
	delete tmp_graph;
	tmp_graph = NULL;
}

#ifdef GRAPH_CONTRACTION
inline void GraphBase::contract(int num, node_id* list)
{
	node* r = nodes + list[0];
	node* i;
	arc* a;
	int k;

	nodes[list[0]].is_contracted = 1;
	for (k=1; k<num; k++)
	{
		nodes[list[k]].is_contracted = 1;
		nodes[list[k]].next_noncontracted = 0;
		UPDATE_EXCESS(nodes + list[0], nodes[list[k]].excess);
	}

	int delta_arc_num = 0;
	int rnum = 0, rnum_orig = (int) (r->last - r->first);
	arc** rlist = NULL;
	for (k=0; k<num; k++)
	{
		i = nodes + list[k];
		for (a=i->first; a<i->last; a++)
		{
			node* j = a->head;
			if (j->is_contracted) { delta_arc_num --; continue; }
			arc* a_rev = REV_ARC(a);
			if (j->a_contraction == NULL)
			{
				j->a_contraction = a;
				a_rev -> head = r;
				rlist = (arc**) common_rbuf->Realloc((rnum+1)*sizeof(arc*));
				rlist[rnum ++] = a;
			}
			else
			{
				delta_arc_num -= 2;
				j->a_contraction          -> r_cap += a -> r_cap;
				REV_ARC(j->a_contraction) -> r_cap += a_rev -> r_cap;
				arc* a_to_move = -- j->last;
				if (a_to_move != a_rev)
				{
#ifdef CONTRACT_MOVE_ARC
					CONTRACT_MOVE_ARC(j, a_to_move, a_rev);
#endif
					REV_ARC(a_to_move)->_rev = ARC_WITH_RCAP(a_rev, a_to_move->r_cap);
					*a_rev = *a_to_move;
				}
			}
		}
		i->last = i->first;
	}

	if (rnum > rnum_orig)
	{
		r->first = (arc*) buf->AllocAligned(rnum*sizeof(arc), 2);
	}
	r->last = r->first;
	for (k=0; k<rnum; k++)
	{
		a = rlist[k];
		a->head->a_contraction = NULL;
		arc* a_rev = REV_ARC(a);
		a_rev->_rev = r->last;
#ifdef KEEP_SATURATION_FLAG
		a->_rev     = ARC_WITH_RCAP(a->_rev,     a_rev->r_cap);
		a_rev->_rev = ARC_WITH_RCAP(a_rev->_rev, a->r_cap);
#endif
		*(r->last ++) = *a;
	}
	r->is_contracted = 0;

	node_num_noncontracted -= num-1;
	edge_num += delta_arc_num / 2;
}

#ifndef IS_GOOD_NODE
#define IS_GOOD_NODE(i) (1)
#endif

#ifdef KEEP_STLISTS

inline int GraphBase::get_smaller_component(int& num, int*& list, ReusableBuffer* rbuf)
{
	int k;
	int n[2] = { 0, 0 }, ngood[2] = { 0, 0 };
	int snum_new[2] = { 0, 0 };
	bool first_phase[2] = { true, true };
	while ( 1 )
	{
		int s = (ngood[0] < ngood[1]) ? 0 : 1;
		if (first_phase[s])
		{
			if (n[s] == snum[s])
			{
				first_phase[s] = false;
				snum_new[s] = snum[s];
				n[s] = 0;
				continue;
			}
			node* i = slist[s][n[s]];
			if (
#ifdef GRAPH_CONTRACTION
				!i->is_contracted && 
#endif
				((s ? (-i->excess) : (i->excess)) > 0))
			{
				n[s] ++;
				if (IS_GOOD_NODE(i)) ngood[s] ++;
			}
			else
			{
				if (s) i->is_in_sink_list = 0;
				else   i->is_in_source_list = 0;
				slist[s][n[s]] = slist[s][-- snum[s]];
			}
		}
		else
		{
			if (n[s] == snum[s])
			{
				num = snum[s];
				list = (int*) rbuf->Alloc(num*sizeof(int));
				for (k=0; k<num; k++) list[k] = (int)(slist[s][k] - nodes);
				if (!first_phase[0])
				{
					for (k=snum_new[0]; k<snum[0]; k++) slist[0][k]->is_in_source_list = 0;
					snum[0] = snum_new[0];
				}
				if (!first_phase[1])
				{
					for (k=snum_new[1]; k<snum[1]; k++) slist[1][k]->is_in_sink_list = 0;
					snum[1] = snum_new[1];
				}
//				if ( 0 )
//				{
//					FOR_ALL_NODES(k) nodes[k].segment = 1-s;
//					for (k=0; k<num; k++) slist[s][k]->segment = s;
//				}
				return s;
			}
			node* i = slist[s][n[s] ++];
			arc* a;
			if (s == 0)
			{
				FOR_ALL_ARCS(i, a)
				if (IS_NONSATURATED(a))
				{
					node* j = a->head;
					if (!j->is_in_source_list)
					{
						j->is_in_source_list = 1; 
						ADD_TO_SLIST(0, j);
						if (IS_GOOD_NODE(j)) ngood[s] ++;
					}
				}
			}
			else
			{
				FOR_ALL_ARCS(i, a)
				if (IS_REV_NONSATURATED(a))
				{
					node* j = a->head;
					if (!j->is_in_sink_list)
					{
						j->is_in_sink_list = 1;
						ADD_TO_SLIST(1, j);
						if (IS_GOOD_NODE(j)) ngood[1] ++;
					}
				}
			}
		}
	}
}
#else
inline int GraphBase::get_smaller_component(int& num, int*& list, ReusableBuffer* rbuf)
{
	printf("error: get_smaller_component() needs KEEP_STLISTS\n");
	exit(1);
}

#endif

inline int GraphBase::get_smaller_balanced_component(int& num, int*& list, ReusableBuffer* rbuf)
{
	set_segment_flags();
	int n[3] = { 0, 0, 0 }, ngood[3] = { 0, 0, 0 };
	int _i;
	FOR_ALL_NODES(_i)
	{
		node* i = nodes + _i;
		n[i->segment] ++;
		if (IS_GOOD_NODE(i)) ngood[i->segment] ++;
	}
	// 2 possible cuts: 
	//     (ngood[0], ngood[1]+ngood[2])
	//     (ngood[0]+ngood[2], ngood[1])
	int a = ngood[0] - (ngood[1]+ngood[2]); int aa = (a > 0) ? a : -a;
	int b = ngood[1] - (ngood[0]+ngood[2]); int bb = (b > 0) ? b : -b;
	int s, s2;
	if (aa < bb)
	{
		if (a > 0) { s = 1; s2 = 2; }
		else       { s = s2 = 0; }
	}
	else
	{
		if (b > 0) { s = 0; s2 = 2; }
		else       { s = s2 = 1; }
	}

	num = 0;
	FOR_ALL_NODES(_i)
	{
		if (nodes[_i].segment == s || nodes[_i].segment == s2)
		{
			list = (int*) rbuf->Realloc((num+1)*sizeof(int));
			list[num ++] = _i;
		}
	}
	return s;
}



#endif


inline void GraphBase::set_segment_flags()
{
	int _i;
	ReusableBufferQueue<node*> q(common_rbuf);
	FOR_ALL_NODES(_i)
	{
		node* i = nodes + _i;
		if (i->excess == 0)
		{
			i->segment = 2;
		}
		else
		{
			i->segment = (i->excess > 0) ? 0 : 1;
			q.add(i);
		}
	}
	while ( q.size() > 0 )
	{
		node* i = q.remove();
		arc* a;
		if (i->segment == 0)
		{
			FOR_ALL_ARCS(i, a)
			if (IS_NONSATURATED(a))
			{
				node* j = a->head;
				if (j->segment == 1) { printf("Error: non-maximal flow!\n"); exit(1); }
				if (j->segment == 2)
				{
					j->segment = 0;
					q.add(j);
				}
			}
		}
		else
		{
			FOR_ALL_ARCS(i, a)
			if (IS_REV_NONSATURATED(a))
			{
				node* j = a->head;
				if (j->segment == 2)
				{
					j->segment = 1;
					q.add(j);
				}
			}
		}
	}
}

inline void GraphBase::print()
{
	int i;
	printf("graph:");
	for (i=0; i<node_num; i++)
	{
#ifdef GRAPH_CONTRACTION
		if (nodes[i].is_contracted) { std::cout << "x "; continue; }
#endif
		std::cout << nodes[i].excess << " ";
	}
	std::cout << "\n      ";
	for (i=0; i<node_num; i++)
	{
		arc* a;
		FOR_ALL_ARCS(nodes+i, a)
		{
			std::cout << (int)(REV_ARC(a)->head - nodes) << "->" << (int)(a->head - nodes) << "[" << GET_RCAP(a) << "] ";
		}
	}
	std::cout << "\n";
}

