#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "block.h"
#include "run_maxflow.h"
#include "timer.h"

#define GRAPH_BASE_REAL int

#ifndef AL_PLUS
#ifndef AL
#ifndef FS_PLUS
#ifndef FS
	#define FS_PLUS
#endif
#endif
#endif
#endif

#ifndef BK
#ifndef IBFS
	#define BK
#endif
#endif

//////////////////////////////////////////////////////////////////
#ifdef AL_PLUS
	#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
#endif
#ifdef AL
	#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
	#define KEEP_REV_ARC_POINTER
#endif
#ifdef FS_PLUS
	#define GRAPH_BASE_FILE "graph_base_forward_star.h"
#endif
#ifdef FS
	#define GRAPH_BASE_FILE "graph_base_forward_star.h"
	#define DO_NOT_KEEP_SATURATION_FLAG
#endif
//////////////////////////////////////////////////////////////////


#ifdef BK

	#define GRAPH_NAMESPACE MAXFLOW_BK
	#include "bk.inc"
	#undef GRAPH_NAMESPACE
	typedef MAXFLOW_BK::Graph G;

#else

	#define GRAPH_NAMESPACE MAXFLOW_IBFS
	#include "ibfs.inc"
	#undef GRAPH_NAMESPACE
	typedef MAXFLOW_IBFS::IBFSGraph G;

#endif



class MaxflowGraph
{
public:
	MaxflowGraph(int node_num);
	~MaxflowGraph();

	void AddEdge(int _i, int _j, GRAPH_BASE_REAL w);

	void MergeParallelEdges();

	void Print()
	{
		Node* i;
		Arc* a;
		for (i=nodes; i<nodes+node_num; i++)
		for (a=i->first; a; a=a->next)
		if (a->head > i)
		{
			printf("%d-%d[%d] ", (int)(i-nodes), (int)(a->head-nodes), (int)a->w);
		}
		printf("\n");
	}

//protected:

	struct Arc;
	struct Node
	{
		Arc*	first;
		Arc*	a_contraction;
	};
	struct Arc
	{
		Node*	head;
		Arc*	next;
		Arc*	rev;
		GRAPH_BASE_REAL	w;
	};

	int		node_num;
	Node*	nodes;

	Buffer	buf;
	ReusableBuffer	rbuf;
};

inline MaxflowGraph::MaxflowGraph(int _node_num)
	: node_num(_node_num), buf(1024)
{
	nodes = new Node[node_num];
	memset(nodes, 0, node_num*sizeof(Node));
}

inline MaxflowGraph::~MaxflowGraph()
{
	delete [] nodes;
}

inline void MaxflowGraph::AddEdge(int _i, int _j, GRAPH_BASE_REAL w)
{
	Node* i = nodes + _i;
	Node* j = nodes + _j;
	Arc* a = (Arc*) buf.Alloc(2*sizeof(Arc));
	Arc* a_rev = a+1;
	a->rev = a_rev; a_rev->rev = a;
	a->w = w;
	a_rev->w = 0;
	a->head = j;
	a_rev->head = i;
	a->next = i->first;
	i->first = a;
	a_rev->next = j->first;
	j->first = a_rev;
}

inline void MaxflowGraph::MergeParallelEdges()
{
	Node* i;
	Arc* a;
	for (i=nodes; i<nodes+node_num; i++)
	{
		Arc** a_prev_ptr = &i->first;
		while ((a=*a_prev_ptr))
		{
			if (a->rev && a->head->a_contraction == NULL)
			{
				a->head->a_contraction = a;
				a_prev_ptr = &a->next;
			}
			else
			{
				if (a->rev)
				{
					a->head->a_contraction->w += a->w;
					a->head->a_contraction->rev->w += a->rev->w;
					a->rev->rev = NULL;
				}
				*a_prev_ptr = a->next;
			}
		}
		for (a=i->first; a; a=a->next)
		{
			a->head->a_contraction = NULL;
		}
	}
}

void RunMaxflow(TmpGraph<int>* g, int s, int t)
{
	MaxflowGraph* M = new MaxflowGraph(g->n);
	
	int e;
	for (e=0; e<g->m; e++) M->AddEdge(g->edges[2*e], g->edges[2*e+1], g->weights[e]);
	M->MergeParallelEdges();
	MaxflowGraph::Node* i;
	MaxflowGraph::Arc* a;
	int m = 0;
	for (i=M->nodes; i<M->nodes+M->node_num; i++)
	{
		if (i == M->nodes+s || i == M->nodes+t) continue;
		for (a=i->first; a; a=a->next)
		{
			MaxflowGraph::Node* j = a->head;
			if (j > i && j != M->nodes+s && j != M->nodes+t) m ++;
		}
	}

	Buffer buf(1024);
	ReusableBuffer rbuf_common;
	ReusableBuffer rbuf[MAXFLOW_RBUF_NUM];
	G h(M->node_num, m, &buf, &rbuf_common, rbuf);
	
	for (i=M->nodes; i<M->nodes+M->node_num; i++)
	{
		if (i == M->nodes+s || i == M->nodes+t) continue;
		for (a=i->first; a; a=a->next)
		{
			MaxflowGraph::Node* j = a->head;
			if (j == M->nodes + s || j == M->nodes + t)
			{
				UPDATE_EXCESS_FIRST(&h, (int)(i-M->nodes), a->rev->w, a->w);
			}
			else if (j > i)
			{
				h.add_edge((int)(i-M->nodes), (int)(j-M->nodes), a->w, a->rev->w);
			}
		}
	}
	GRAPH_BASE_REAL flow;
	double T_start = get_time();
	MAXFLOW_SINGLE(&h, flow);
	printf("flow=%d\n", (int) flow);
	printf("time=%f secs\n", get_time() - T_start);
}

